﻿using System;
using System.Collections.Specialized;
using System.Collections.Generic;
using System.Configuration;
using System.Data.Caching;

using NorthwindCacheDemo.DTO;
using NorthwindCacheDemo.DAL;

namespace NorthwindCacheDemo.CL.Velocity
{
    /// <summary>
    /// Provides a cache layer for the Northwind product catalog.
    /// VelocityCatalogCache acts as a read-through layer in which
    /// operations that result in a cache miss will subsequently 
    /// populate the cache by generating calls to the underlying
    /// data store. Methods in this class first attempt to retrieve
    /// products and categories from the catalog and reports named
    /// caches. If these methods result in a cache miss, they will
    /// call into the data layer to populate the cache and provide
    /// the caller with the result. These methods do not provide
    /// concurrency support. As this data is primarily read-only,
    /// the impact will not be too dramatic; however, for a more
    /// robust approach to concurrency, these methods could take
    /// locks before going to the data tier if a cache miss occurs.
    /// For more information on concurrency, please see the
    /// documentation provided with the Velocity installation.
    /// </summary>
    public class VelocityCatalogCache : ICatalogCache
    {
        private const string CATEGORY_KEY_FORMAT = "Category-{0}";
        private const string PRODUCT_KEY_FORMAT = "Product-{0}";
        private const string PRODUCT_LIST_KEY_FORMAT = "ProductList-{0}";
        private const string POPULAR_LIST_KEY_FORMAT = "PopularProductList-{0}";
        private const string ALSO_PURCHASED_KEY_FORMAT = "AlsoPurchasedProductList-{0}";
        private const string CATEGORY_LIST_KEY = "CategoryList";
        private const string POPULAR_LIST_KEY = "PopularProductList";

        private Cache _catalogCache;

        private Cache _reportCache;

        public VelocityCatalogCache()
        {
            NameValueCollection config = (NameValueCollection)ConfigurationManager.GetSection("northwindCacheDemo");

            // Create end point for local caching
            ServerEndPoint endPoint = new ServerEndPoint(
                config["VelocityHostName"], 
                Convert.ToInt32(config["VelocityPort"]), 
                config["VelocityCacheHostName"]);

            // Create catalog cache client with local caching
            CacheFactory factory = new CacheFactory(new ServerEndPoint[] { endPoint }, true, true);
            _catalogCache = factory.GetCache("catalog");

            // Create reports cache client without local caching
            factory = new CacheFactory();
            _reportCache = factory.GetCache("reports");
        }

        public IList<Category> GetCategories()
        {
            IList<int> categoryIDs = (IList<int>)_catalogCache.Get(CATEGORY_LIST_KEY);
            IList<Category> categories = null;

            if (categoryIDs != null)
            {
                categories = new List<Category>(categoryIDs.Count);
                foreach (int id in categoryIDs)
                {
                    categories.Add((Category)_catalogCache.Get(String.Format(CATEGORY_KEY_FORMAT, id)));                    
                }
            }
            else
            {
                categories = DataLayer.CatalogProvider.GetCategories();

                if (categories != null && categories.Count > 0)
                {
                    categoryIDs = new List<int>(categories.Count);
                    foreach (Category c in categories)
                    {
                        categoryIDs.Add(c.CategoryID);
                        _catalogCache.Put(String.Format(CATEGORY_KEY_FORMAT, c.CategoryID), c);
                    }

                    _catalogCache.Put(CATEGORY_LIST_KEY, categoryIDs);
                }
            }

            return categories;
        }

        public Category GetCategory(int id)
        {
            string key = String.Format(CATEGORY_KEY_FORMAT, id);
            Category c = (Category)_catalogCache.Get(key);

            if (c == null)
            {
                c = DataLayer.CatalogProvider.GetCategory(id);

                if (c != null)
                {
                    _catalogCache.Put(key, c);
                }
            }

            return c;
        }

        public Product GetProduct(int id)
        {
            string key = String.Format(PRODUCT_KEY_FORMAT, id);
            Product p = (Product)_catalogCache.Get(key);

            if (p == null)
            {
                p = DataLayer.CatalogProvider.GetProduct(id);

                if (p != null)
                {
                    _catalogCache.Put(key, p);
                }
            }

            return p;
        }

        public IList<Product> GetProductsByCategory(int id)
        {
            string listKey = String.Format(PRODUCT_LIST_KEY_FORMAT, id);

            IList<int> productIDs = (IList<int>)_catalogCache.Get(listKey);
            IList<Product> products = null;

            if (productIDs != null)
            {
                products = new List<Product>(productIDs.Count);

                foreach (int productID in productIDs)
                {
                    products.Add((Product)_catalogCache.Get(String.Format(PRODUCT_KEY_FORMAT, productID)));
                }
            }
            else
            {
                products = DataLayer.CatalogProvider.GetProductsByCategory(id);

                if (products != null && products.Count > 0)
                {
                    productIDs = new List<int>(products.Count);
                    foreach (Product p in products)
                    {
                        productIDs.Add(p.ProductID);

                        _catalogCache.Put(String.Format(PRODUCT_KEY_FORMAT, p.ProductID), p);
                    }

                    _catalogCache.Put(listKey, productIDs);
                }
            }

            return products;
        }

        public IList<Product> GetPopularProducts()
        {
            IList<int> productIDs = (IList<int>)_reportCache.Get(POPULAR_LIST_KEY);
            IList<Product> products = null;

            if (productIDs != null)
            {
                products = new List<Product>(productIDs.Count);

                foreach (int productID in productIDs)
                {
                    products.Add((Product)_catalogCache.Get(String.Format(PRODUCT_KEY_FORMAT, productID)));
                }
            }
            else
            {
                products = DataLayer.CatalogProvider.GetPopularProducts();

                if (products != null && products.Count > 0)
                {
                    productIDs = new List<int>(products.Count);
                    foreach (Product p in products)
                    {
                        productIDs.Add(p.ProductID);

                        _catalogCache.Put(String.Format(PRODUCT_KEY_FORMAT, p.ProductID), p);
                    }

                    _reportCache.Put(POPULAR_LIST_KEY, productIDs);
                }
            }

            return products;
        }

        public IList<Product> GetPopularProductsByCategory(int id)
        {
            string listKey = String.Format(POPULAR_LIST_KEY_FORMAT, id);

            IList<int> productIDs = (IList<int>)_reportCache.Get(listKey);
            IList<Product> products = null;

            if (productIDs != null)
            {
                products = new List<Product>(productIDs.Count);
                
                foreach (int productID in productIDs)
                {
                    products.Add((Product)_catalogCache.Get(String.Format(PRODUCT_KEY_FORMAT, productID)));
                }
            }
            else
            {
                products = DataLayer.CatalogProvider.GetPopularProductsByCategory(id);

                if (products != null && products.Count > 0)
                {
                    productIDs = new List<int>(products.Count);
                    foreach (Product p in products)
                    {
                        productIDs.Add(p.ProductID);
                        _catalogCache.Put(String.Format(PRODUCT_KEY_FORMAT, p.ProductID), p);
                    }

                    _reportCache.Put(listKey, productIDs);
                }
            }

            return products;
        }

        public IList<Product> GetAlsoPurchasedProducts(int id)
        {
            string listKey = String.Format(ALSO_PURCHASED_KEY_FORMAT, id);

            IList<int> productIDs = (IList<int>)_reportCache.Get(listKey);
            IList<Product> products = null;

            if (productIDs != null)
            {
                products = new List<Product>(productIDs.Count);

                foreach (int productID in productIDs)
                {
                    products.Add((Product)_catalogCache.Get(String.Format(PRODUCT_KEY_FORMAT, productID)));
                }
            }
            else
            {
                products = DataLayer.CatalogProvider.GetAlsoPurchasedProducts(id);

                if (products != null && products.Count > 0)
                {
                    productIDs = new List<int>(products.Count);
                    foreach (Product p in products)
                    {
                        productIDs.Add(p.ProductID);

                        _catalogCache.Put(String.Format(PRODUCT_KEY_FORMAT, p.ProductID), p);
                    }

                    _reportCache.Put(listKey, productIDs);
                }
            }

            return products;
        }
    }
}
