﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Data.SqlClient;

using NorthwindCacheDemo.DAL;
using NorthwindCacheDemo.DTO;

namespace NorthwindCacheDemo.DAL.SqlClient
{
    /// <summary>
    /// Provides data access to the Northwind product catalog.
    /// The methods in this class use System.Data.SqlClient to 
    /// retrieve data from the Northwind database. Further, by using 
    /// the SqlDataReader, these data access methods build transfer 
    /// objects that are subsequently passed to consuming layers.
    /// </summary>
    public class SqlCatalogProvider : SqlProvider, ICatalogProvider
    {
        public SqlCatalogProvider() { }

        public IList<Category> GetCategories()
        {
            IList<Category> categories = null;
            using (SqlConnection conn = new SqlConnection(this.ConnectionString))
            {
                SqlCommand cmd = new SqlCommand("SELECT * FROM [Northwind].[dbo].[Categories]", conn);
                cmd.CommandType = CommandType.Text;
                conn.Open();

                categories = CreateCategoriesFromDataReader(cmd.ExecuteReader());
            }

            return categories;
        }

        public Category GetCategory(int id)
        {
            Category category = null;

            using (SqlConnection conn = new SqlConnection(this.ConnectionString))
            {
                SqlCommand cmd = new SqlCommand("SELECT * FROM [Northwind].[dbo].[Categories] AS c WHERE c.CategoryID = " + id.ToString(), conn);
                cmd.CommandType = CommandType.Text;
                conn.Open();

                SqlDataReader dr = cmd.ExecuteReader();

                if (!dr.Read())
                {
                    return null;
                }

                category = CreateCategoryFromDataReader(dr);
            }

            return category;
        }

        public Product GetProduct(int id)
        {
            Product product = null;

            using (SqlConnection conn = new SqlConnection(this.ConnectionString))
            {
                SqlCommand cmd = new SqlCommand("SELECT * FROM [Northwind].[dbo].[Products] AS p WHERE p.ProductID = " + id.ToString(), conn);
                cmd.CommandType = CommandType.Text;
                conn.Open();

                SqlDataReader dr = cmd.ExecuteReader();

                if (!dr.Read())
                {
                    return null;
                }

                product = CreateProductFromDataReader(dr);
            }

            return product;
        }

        public IList<Product> GetProductsByCategory(int id)
        {
            IList<Product> products = null;

            using (SqlConnection conn = new SqlConnection(this.ConnectionString))
            {
                SqlCommand cmd = new SqlCommand("SELECT * FROM [Northwind].[dbo].[Products] AS p WHERE p.CategoryID = " + id.ToString(), conn);
                cmd.CommandType = CommandType.Text;
                conn.Open();

                products = CreateProductsFromDataReader(cmd.ExecuteReader());
            }

            return products;
        }

        public IList<Product> GetPopularProducts()
        {
            string commandText = 
                @"SELECT * 
                FROM [Northwind].[dbo].[Products] AS p
                WHERE p.ProductID IN
	                (SELECT TopProducts.ProductID 
	                FROM
		                (SELECT TOP 10 items.ProductID AS ProductID, COUNT(items.ProductID) AS ProductCount
		                FROM [Northwind].[dbo].[Order Details] AS items
		                GROUP BY items.ProductID
		                ORDER BY ProductCount DESC) AS TopProducts)";
            
            IList<Product> products = null;

            using (SqlConnection conn = new SqlConnection(this.ConnectionString))
            {
                SqlCommand cmd = new SqlCommand(commandText, conn);
                cmd.CommandType = CommandType.Text;
                conn.Open();

                products = CreateProductsFromDataReader(cmd.ExecuteReader());
            }

            return products;
        }

        public IList<Product> GetPopularProductsByCategory(int id)
        {
            string commandText =
                @"SELECT * 
                FROM [Northwind].[dbo].[Products] AS p
                WHERE p.ProductID IN
	                (SELECT TopProducts.ProductID 
	                FROM
		                (SELECT TOP 3 items.ProductID AS ProductID, COUNT(items.ProductID) AS ProductCount
		                FROM [Northwind].[dbo].[Order Details] AS items
		                WHERE items.ProductID IN 
			                (SELECT categoryProducts.ProductID
			                FROM [Northwind].[dbo].[Products] AS categoryProducts
			                WHERE categoryProducts.CategoryID = " + id.ToString() + @")
		                GROUP BY items.ProductID
		                ORDER BY ProductCount DESC) AS TopProducts)";

            IList<Product> products = null;

            using (SqlConnection conn = new SqlConnection(this.ConnectionString))
            {
                SqlCommand cmd = new SqlCommand(commandText, conn);
                cmd.CommandType = CommandType.Text;
                conn.Open();

                products = CreateProductsFromDataReader(cmd.ExecuteReader());
            }

            return products;
        }

        public IList<Product> GetAlsoPurchasedProducts(int id)
        {
            string idString = id.ToString();

            string commandText = 
                @"SELECT *
                FROM [Northwind].[dbo].[Products] AS p
                WHERE p.ProductID IN 
	                (SELECT TopProducts.ProductID
	                FROM
		                (SELECT TOP 5 i.ProductID AS ProductID, COUNT(i.ProductID) AS ProductCount
		                FROM [Northwind].[dbo].[Order Details] as i
		                WHERE i.OrderID IN
			                (SELECT d.OrderID
			                FROM [Northwind].[dbo].[Order Details] as d
			                WHERE d.ProductID = " + idString + @")
		                GROUP BY i.ProductID
		                HAVING i.ProductID <> " + idString + @"
		                ORDER BY ProductCount DESC) AS TopProducts)";

            IList<Product> products = null;

            using (SqlConnection conn = new SqlConnection(this.ConnectionString))
            {
                SqlCommand cmd = new SqlCommand(commandText, conn);
                cmd.CommandType = CommandType.Text;
                conn.Open();

                products = CreateProductsFromDataReader(cmd.ExecuteReader());
            }

            return products;
        }

        private IList<Category> CreateCategoriesFromDataReader(SqlDataReader dr)
        {
            IList<Category> categories = new List<Category>();
            while (dr.Read())
            {
                categories.Add(CreateCategoryFromDataReader(dr));
            }
            return (categories.Count > 0) ? categories : null;
        }

        private Category CreateCategoryFromDataReader(SqlDataReader dr)
        {
            return new Category()
            {
                CategoryID = Convert.ToInt32(dr["CategoryID"]),
                Name = Convert.ToString(dr["CategoryName"]),
                Description = Convert.ToString(dr["Description"])
            };
        }

        private IList<Product> CreateProductsFromDataReader(SqlDataReader dr)
        {
            IList<Product> products = new List<Product>();
            while (dr.Read())
            {
                products.Add(CreateProductFromDataReader(dr));
            }
            return (products.Count > 0) ? products : null;
        }

        private Product CreateProductFromDataReader(SqlDataReader dr)
        {
            return new Product()
            {
                ProductID = Convert.ToInt32(dr["ProductID"]),
                CategoryID = Convert.ToInt32(dr["CategoryID"]),
                Name = Convert.ToString(dr["ProductName"]),
                QuantityPerUnit = Convert.ToString(dr["QuantityPerUnit"]),
                UnitPrice = Convert.ToDecimal(dr["UnitPrice"])
            };
        }
    }
}
